#!/usr/bin/env python3
#R exposed top routines
import pickle, re, sys, operator, LSD
import pandas as pd, numpy as np
from collections import OrderedDict
import matplotlib.pyplot as plt
from rpy2.rinterface import RRuntimeError
from rpy2.robjects.packages import importr
import rpy2.robjects as ro
#Activate automatic pandas/r conversion
from rpy2.robjects import pandas2ri
pandas2ri.activate()

# General importr's
base = importr('base')
stats = importr('stats')
utils = importr('utils')

# Gene-set enrichment analysis
def get_gcindices_r(countsGeneLabels,correctBackground=False,remove_empty=True):
    """
    >>> indices_r = get_gcindices(counts.index,correctBackground=False) # doctest: +SKIP
    """
    limma = importr('limma')
    gc = LSD.get_msigdb6()

    if correctBackground:
        gc = {gsc:{gs:[g for g in gc[gsc][gs] if g in countsGeneLabels] for gs in gc[gsc]} for gsc in gc}
        
    countsGeneLabels_r = ro.StrVector(countsGeneLabels)        
    gc_indices_r = {gsc:limma.ids2indices(ro.ListVector(gc[gsc]),countsGeneLabels_r,remove_empty=remove_empty)
                    for gsc in gc}
    
    return gc_indices_r

def romer(counts_r,index_r,design_r,contrast_r):
    """
    counts_r -> should be log-expression values (voom transformed RNAseq counts also good)
    index_r -> collection of geneses (ListVector of IntVector's)
    design_r -> design matrix
    contrast_r -> coef or contrast vector
      e.g. contrasts_r.rx(True,"celllineIMR32.shairpinsh25+celllineIMR32.shairpinsh27"

    >>> romer(counts_r,index_r,design_r,contrast_r) # doctest: +SKIP
    """
    limma = importr('limma')
    rr = limma.romer(counts_r,index=index_r,design=design_r,contrast=contrast_r)
    rr_py = pd.DataFrame(pandas2ri.ri2py(rr),columns=base.colnames(rr),index=base.rownames(rr))
    rr_py['Up_adj'] = stats.p_adjust(rr.rx(True,'Up'),method='fdr')
    rr_py['Down_adj'] = stats.p_adjust(rr.rx(True,'Down'),method='fdr')
    rr_py['min_p_adj'] = rr_py[['Up_adj','Down_adj']].T.min()
    rr_py.sort_values('min_p_adj',inplace=True)
    return rr_py

def romer_fullMSigDB(counts_r,countsGeneLabels,**kwargs):
    """
    >>> romer_fullMSigDB(counts_r,countsGeneLabels) # doctest: +SKIP
    """
    gc = get_gcindices_r(countsGeneLabels)
    return OrderedDict([
        (gsc,romer(counts_r,index_r=gc[gsc],**kwargs))
        for gsc in sorted(gc)
    ])

def mroast(counts_r,index_r,design_r,contrast_r,set_statistic="mean"):
    """
    >>> mroast(counts_r,index_r,design_r,contrast_r) # doctest: +SKIP
    """
    limma = importr('limma')
    return base.data_frame(limma.mroast(counts_r,index_r,design=design_r,contrast=contrast_r,set_statistic=set_statistic))

def genesets2indices_r(genesets,countsGeneLabels):
    """
    genesets should be a dictionary of genesets
    >>> from .tests.test_retro import testGenesets, testCountsGenelabels
    >>> print(genesets2indices_r(testGenesets,testCountsGenelabels))
    [[1]]
    [1] 1 2
    <BLANKLINE>
    [[2]]
    [1] 2 3
    <BLANKLINE>
    <BLANKLINE>
    """
    limma = importr('limma')    
    genesets_r = ro.ListVector(genesets)
    return limma.ids2indices(genesets, countsGeneLabels)

def barcodeplot(fit_r,contrast,indices,geneset,geneset2=None,pngname=None,width=512,height=512,**kwargs):
    """
    >>> barcodeplot(fit_r,3,indices['C5'],'GO_GLUTATHIONE_TRANSFERASE_ACTIVITY',pngname=f.name) # doctest: +SKIP
    """
    limma = importr('limma')
    if pngname:
        grdevices = importr('grDevices')
        grdevices.png(file=pngname, width=width, height=height)
    limma.barcodeplot(fit_r.rx2('t').rx(True,contrast),indices.rx2(geneset),
                      index2 = indices.rx2(geneset2) if geneset2 else ro.NULL,**kwargs)
    if pngname: grdevices.dev_off()

def rankSumProbsMSigDB(ranks,universe,adjpmethod='fdr'):
    from bidali.fegnome import RankSumResult,rankSumProbability
    gc = LSD.get_msigdb6()
    results = OrderedDict()
    for gsc in sorted(gc):
        results[gsc] = pd.DataFrame({gs:rankSumProbability(ranks,gc[gsc][gs]) for gs in gc[gsc]}).T
        results[gsc].columns = RankSumResult._fields
        results[gsc]['fepa'] = stats.p_adjust(results[gsc].fe_p,method=adjpmethod)
        results[gsc]['respa'] = stats.p_adjust(results[gsc].probability,method=adjpmethod)
        results[gsc].sort_values(['fepa','respa'],inplace=True)

    return results



import sys
from urllib import error,parse
import urllib.request
import json
import time

class EnsemblRestClient:
    def __init__(self, server='http://rest.ensembl.org', reqs_per_sec=15):
        self.server = server
        self.reqs_per_sec = reqs_per_sec
        self.req_count = 0
        self.last_req = 0

    def perform_rest_action(self, endpoint, headers=None, params=None):
        if headers is None:
            headers = {}

        if 'Content-Type' not in headers:
            headers['Content-Type'] = 'application/json'

        if params:
            endpoint += '?' + parse.urlencode(params)

        data = None
        if self.req_count >= self.reqs_per_sec: # check if we need to rate limit ourselves
            delta = time.time() - self.last_req
            if delta < 1:
                time.sleep(1 - delta)
            self.last_req = time.time()
            self.req_count = 0
        try:
            request = urllib.request.Request(self.server + endpoint, headers=headers)
            response = urllib.request.urlopen(request)
            content = response.read()
            if content:
                data = json.loads(content.decode())
            self.req_count += 1
        except error.HTTPError as e:
            if e.code == 429: # check if we are being rate limited by the server
                if 'Retry-After' in e.headers:
                    retry = e.headers['Retry-After']
                    time.sleep(float(retry))
                    self.perform_rest_action(endpoint, headers, params)
            else:
                sys.stderr.write('Request failed for {0}: Status code: {1.code} Reason: {1.reason}\n'.format(endpoint, e))
        return data
    
    def get_variants(self, species, symbol):
        genes = self.perform_rest_action(
            '/xrefs/symbol/{0}/{1}'.format(species, symbol), 
            params={'object_type': 'gene'}
        )
        if genes:
            stable_id = genes[0]['id']
            variants = self.perform_rest_action(
                '/overlap/id/{0}'.format(stable_id),
                params={'feature': 'variation'}
            )
            return variants
        return None

def run(species, symbol):
    """
    Example:
    run('human', 'BRAF')
    """
    client = EnsemblRestClient()
    variants = client.get_variants(species, symbol)
    if variants:
        for v in variants:
            print('{seq_region_name}:{start}-{end}:{strand} ==> {id} ({consequence_type})'.format(**v))

def getSequences(geneset,file,expand_5prime=0,sequenceEnd=-1):
    """
    getSequences expects a list of gene symbols, to retrieve sequences
    and a file name, to which the sequences are stored for future
    requests
    expand_5prime: include x bp upstream
    sequenceEnd: only keep x bp; -1 -> whole sequence

    Raises exception if symbols in file do not match requested symbols
    TODO also raise if other params do not match
    """
    #Settings
    currentSettings = (sorted(geneset),expand_5prime,sequenceEnd)
    
    #Imports
    import pickle

    #Setup client
    client = EnsemblRestClient()

    #Get sequences
    try:
        settings,sequences = pickle.load(open(file,'rb'))
        if settings != currentSettings:
            raise Exception('File content and settings provided do not match')
    except FileNotFoundError:
        sequences = {}
        for gene in geneset:
            try:
                geneID = client.perform_rest_action('/xrefs/symbol/human/{}'.format(gene))[0]['id']
                sequence = client.perform_rest_action('/sequence/id/{}'.format(geneID),
                                                      params={'expand_5prime':expand_5prime})['seq']
                sequences[gene] = sequence[:sequenceEnd]
            except IndexError: pass
        pickle.dump((currentSettings,sequences),open(file,'wb'))
    return sequences

def getOrthologs(symbol,species='human',ospecies='mus_musculus'):
    #Setup client
    client = EnsemblRestClient()
    
    data = client.perform_rest_action('/homology/symbol/{}/{}'.format(
        species,parse.quote(symbol)))
    retrieved_homologies = []
    for homologies in data['data']:
        id = homologies['id']
        for homology in homologies['homologies']:
            if ospecies in homology['target']['species']:
                retrieved_homologies.append(homology)
    osymbols = []
    for homology in retrieved_homologies:
        data = client.perform_rest_action('/xrefs/id/{}'.format(
            parse.quote(homology['target']['id'])))
        for db in data:
            if db['dbname'] == 'EntrezGene':
                osymbols.append(db['display_id'])
                break
    return osymbols



from urllib import error,parse
from urllib.request import Request,urlopen
import json

def fetchGenenamesJSON(symbol,alias=False):
    """
    Requires a gene symbol
    """
    if not alias:
        requesturl = 'http://rest.genenames.org/fetch/symbol/{}'
    else:
        requesturl = 'http://rest.genenames.org/fetch/alias_symbol/{}'
    requesturl = requesturl.format(parse.quote(symbol))
    request = Request(requesturl,
                      headers={'Accept':'application/json'})
    response = urlopen(request)
    content = response.read()
    if content:
        return json.loads(content.decode())
    
def fetchAliases(symbol):
    """
    Returns list of all aliases for a symbol.
    First alias in list is official symbol.
    """
    symbolJSON = fetchGenenamesJSON(symbol)
    if not symbolJSON['response']['numFound']:
        symbolJSON = fetchGenenamesJSON(symbol,alias=True)
    symbols = [symbolJSON['response']['docs'][0]['symbol']
    ]+symbolJSON['response']['docs'][0]['alias_symbol']
    return symbols



#!/usr/bin/env python
# LSD: Localised Structured Datasets
# Module that preprocesses commonly used datasets
# and makes them available for use in python.

import gzip, pickle, time, os
from zipfile import ZipFile
from io import TextIOWrapper, StringIO
import pandas as pd, numpy as np
from os.path import expanduser, exists
from collections import OrderedDict
from contextlib import redirect_stdout, redirect_stderr

## Defaults
datadir = expanduser('~/Dropbiz/Lab/z_archive/Datasets/')
processedDataStorage = expanduser('~/Data/LSDpy/')

## Utility functions
def getLSDataset(name,**kwargs):
    try: return name(**kwargs)
    except TypeError:
        return globals()[name](**kwargs)

def listLSDatasets(subpackages=True):
    print(*[i for i in sorted(globals()) if i.startswith('get_')],sep='\n')
    if subpackages:
        print('\nWithin subpackages:')
        import pkgutil as pk
        from inspect import getmembers
        for mod in pk.walk_packages(__path__,__name__+'.'):
            if not mod.ispkg:
                mol = pk.importlib.import_module(mod[1])
                print('>',mod[1])
                print(*[i for i in sorted(dict(getmembers(mol)))
                        if i.startswith('get_')],sep='\n',end='\n\n')

class Dataset:
    """
    A Dataset object is a collection of datasets, accessible as
    attributes from the Dataset object.

    to_R pushes the sub datasets to R making them available in the
    global namespace
    """
    def __init__(self,**kwargs):
        self.__datasets__ = set(kwargs)
        for kw in kwargs:
            self.__setattr__(kw,kwargs[kw])

    def generateRData(self):
        pass
    
class DatasetRepo:
    """
    Object containing:
      - a Dataset object
      - a report on how the dataset was generated
      - file location
      - a history of the code that generated the current
        dataset and earlier versions
      - a hash of the code that generated the current dataset
      - archived earlier dataset objects

    Method to keep only the most recent version of the
    dataset to free up hard disk space
    """
    def __init__(self,dataset,code,report,filename):
        import hashlib, pickle
        self.dataset = dataset
        self.report = report
        self.currentHash = hashlib.md5(code.encode()).hexdigest()
        self.code = OrderedDict([(self.currentHash,code)])
        self.archive = OrderedDict()
        self.filename = filename
        pickle.dump(self,open(filename,'wb'))

    def update(self,dataset,code,report):
        import hashlib, pickle
        # Put previous object in archive
        self.archive[self.currentHash] = self.dataset

        # Update
        self.dataset = dataset
        self.report = report
        self.currentHash = hashlib.md5(code.encode()).hexdigest()
        self.code[self.currentHash] = code
        pickle.dump(self,open(self.filename,'wb'))

    def wipeArchive(self):
        self.archive = OrderedDict()
        pickle.dump(self,open(self.filename,'wb'))
        
def retrieveSources(dataset_getfunction):
    """
    A dataset_getfunction function that contains 'Source:' lines
    in the docstring, can be decorated with this function.
    If a source is not locally available, it will be downloaded
    and added to the processedDataStorage location.

    A source line has to be formatted accordingly:
    Source: [filename] url

    If filename is not provided, the last part of the url (after last '/')
    is taken as filename.

    Source lines can also be provided as arguments to the FileNotFoundError
    that the dataset_getfunction throws.
    """
    import inspect, requests
    from urllib.request import urlopen, urlretrieve

    def wrapper(*args, **kwargs):
        try:
            return dataset_getfunction(*args, **kwargs)
        except FileNotFoundError as fnf:
            for docline in inspect.getdoc(dataset_getfunction).split('\n') + list(fnf.args):
                if docline.startswith('Source:'):
                    docline = docline.split()
                    if len(docline) == 2:
                        url = docline[1]
                        filename = url[url.rindex('/')+1:]
                    elif len(docline) == 3:
                        url = docline[2]
                        filename = docline[1]
                    if not exists(processedDataStorage+filename):
                        print('Downloading {}:'.format(filename))
                        if url.startswith('ftp://'):
                            urlretrieve(url,processedDataStorage+filename,
                                        lambda x,y,z: print("\r[{}{}]".format('=' * int(50*x*y/z),
                                                                              ' ' * (50-int(50*x*y/z))),
                                                            end='',flush=True))
                        else:
                            r = requests.get(url,stream=True)
                            total_length = r.headers.get('content-length')
                            with open(processedDataStorage+filename,'wb') as f:
                                if total_length is None: f.write(r.content)
                                else:
                                    dl = 0
                                    total_length = int(total_length)
                                    for data in r.iter_content(chunk_size=4096):
                                        dl += len(data)
                                        f.write(data)
                                        done = int(50 * dl / total_length)
                                        print("\r[{}{}]".format('=' * done, ' ' * (50-done)),end='',flush=True)
            try: return dataset_getfunction(*args, **kwargs)
            except FileNotFoundError:
                print('Either not all source files are documented correctly in docstring,',
                      'or there is a source file unrelated issue')
                raise

    return wrapper

def storeDatasetLocally(dataset_getfunction):
    """
    Can be used as a decorator for 'get_dataset' functions.
    It will check if a processed dataset is locally available,
    and if so, load that one instead of processing from the source
    files.

    Should only be used for functions that do not process the data
    differently depending on the 'get_dataset' function arguments.
    The wrapper raises a warning if there are arguments to pass to 
    the 'get_dataset' function.
    """
    import inspect, hashlib, re
    from plumbum import colors
    dependency = re.compile(r'\W*Dependenc(y|ies): (.+)')
    
    def wrapper(*args, verbose=True, **kwargs):
        if args or kwargs:
            import warnings
            warnings.warn(
                'This decorated function is not designed to use with arguments. Be warned!'
            )
        # Check if data was already processed
        ## Prepare hash
        functionSource = inspect.getsource(dataset_getfunction).encode()
        ### Check if dependencies
        docstr = inspect.getdoc(dataset_getfunction)
        if docstr:
            dependencies = [e for d in (d.groups()[1].split()
                                        for d in (dependency.search(l)
                                                  for l in docstr.split('\n')) if d)
                            for e in d]
            for d in dependencies:
                d = d.split('.')
                d = getattr(globals()[d[0]],d[-1],globals()[d[0]]) #hack to get d with globals and getattr
                try:
                    functionSource+=inspect.getsource(d).encode()
                except TypeError:
                    functionSource+=pickle.dumps(d)
        hashvalue = hashlib.md5(functionSource).hexdigest()
        datastorage = '{}{}.pickle'.format(processedDataStorage,
                                              dataset_getfunction.__name__.replace('get_','')
        )
        
        if exists(datastorage):
            datasetrepo = pickle.load(open(datastorage,'rb'))
            if datasetrepo.currentHash == hashvalue:
                print(datasetrepo.report)
                if verbose:
                    print(colors.green & colors.bold | 'Repo size {:.1f}MB, archive contains {} other versions'.format(
                        os.stat(datastorage).st_size/1024**2,
                        len(datasetrepo.archive)
                    ))
                return datasetrepo.dataset
            else:
                print(colors.cyan | 'Dataset content out of date, updating:')
                updateRepo = True
        else:
            print(colors.cyan | 'Dataset not locally available, generating:')
            updateRepo = False

        # Redirect stdout and stderr
        stouterr_redirect = StringIO()
        with redirect_stdout(stouterr_redirect), redirect_stderr(stouterr_redirect):
            start = time.process_time()
            # Run the function that generates the dataset
            dataset = dataset_getfunction(*args, **kwargs)
            duration = time.process_time() - start
            print('Dataset',datastorage,'generated',time.strftime('%c'))
            print('Processing time',time.strftime('%H:%M:%S', time.gmtime(duration)))
            
        print(stouterr_redirect.getvalue())

        if updateRepo:
            datasetrepo.update(dataset,functionSource.decode(),stouterr_redirect.getvalue().strip())
        else:
            try: DatasetRepo(dataset,functionSource.decode(),stouterr_redirect.getvalue().strip(),datastorage)
            except FileNotFoundError:
                print('Not possible to store dataset locally. Create',processedDataStorage,
                      'if you want to avoid reprocessing dataset on every call.')
            
        return dataset

    return wrapper

## Datasets
### References/annotations
from .dealer.external.ensembl import get_ensembl, get_ensemblGeneannot

@retrieveSources
def get_entrez():
    """
    Source: ftp://ftp.ncbi.nih.gov/refseq/H_sapiens/RefSeqGene/gene_RefSeqGene
    """
    entrez = pd.read_table(processedDataStorage+'gene_RefSeqGene', index_col='GeneID')
    return entrez

@retrieveSources
def get_liftover(frm=19,to=38):
    """
    Info: http://hgdownload.cse.ucsc.edu/downloads.html
    """
    from pyliftover import LiftOver
    liftoverfile = 'hg{}ToHg{}.over.chain.gz'.format(frm,to)
    try: return LiftOver(processedDataStorage+liftoverfile)
    except FileNotFoundError:
        raise FileNotFoundError('Source: http://hgdownload.cse.ucsc.edu/gbdb/hg{}/liftOver/{}'.format(frm,liftoverfile))

def get_lift19to38():
    """
    Source: http://hgdownload.cse.ucsc.edu/gbdb/hg19/liftOver/hg19ToHg38.over.chain.gz
    """
    return get_liftover(frm=19,to=38)

@storeDatasetLocally
def get_proteinNetworks():
    """
    Source: https://thebiogrid.org/downloads/archives/Release%20Archive/BIOGRID-3.4.147/BIOGRID-ALL-3.4.147.tab2.zip
    Source: http://string-db.org/download/protein.links.v10/9606.protein.links.v10.txt.gz
    Source: http://string-db.org/mapping_files/entrez_mappings/entrez_gene_id.vs.string.v10.28042015.tsv
    """
    import networkx as nx

    #Biogrid
    with ZipFile(datadir+'ProteinNetworks/BIOGRID-ALL-3.4.147.tab2.zip') as biogridzip:
        ds = pd.read_table(TextIOWrapper(biogridzip.open('BIOGRID-ALL-3.4.147.tab2.txt','r')),low_memory=False)
    ds = ds[ds['Organism Interactor A'] == 9606]
    Gbio = nx.Graph()
    ds.T.apply(lambda x: Gbio.add_edge(x['Official Symbol Interactor A'],x['Official Symbol Interactor B']))

    #String-DB
    stringdb = pd.read_table(gzip.open(datadir+'ProteinNetworks/9606.protein.links.v10.txt.gz','rt'),sep=' ')
    stringids = pd.read_table(datadir+'ProteinNetworks/entrez_gene_id.vs.string.v10.28042015.tsv',index_col='STRING_Locus_ID')
    entrez = get_entrez()
    stringids = stringids[stringids['#Entrez_Gene_ID'].isin(entrez.index)]
    stringdb = stringdb[stringdb.combined_score > 400] #study stringdb.combined_score.hist(bins='auto') to set threshold
    stringdb = stringdb[stringdb.protein1.isin(stringids.index) & stringdb.protein2.isin(stringids.index)]
    stringdb.protein1 = stringdb.protein1.apply(lambda x: stringids.ix[x]['#Entrez_Gene_ID'])
    stringdb.protein2 = stringdb.protein2.apply(lambda x: stringids.ix[x]['#Entrez_Gene_ID'])
    stringdb.protein1 = stringdb.protein1.apply(lambda x: entrez.ix[x].Symbol)
    stringdb.protein2 = stringdb.protein2.apply(lambda x: entrez.ix[x].Symbol)
    Gstring = nx.Graph()
    stringdb.T.apply(lambda x: Gstring.add_edge(x.protein1,x.protein2))
    
    return Dataset(biogridnx = Gbio, biogrid = ds,
                   stringnx = Gstring, string = stringdb)
    
@storeDatasetLocally
def get_centromeres():
    """
    Source: from R bioconductor GWASTools: data(centromeres.hg38)
    """
    from bidali.seqanalysis import loadHumanGenome

    # Centromere positions
    centromereshg38="""1      1 122026460  125184587
2      2  92188146   94090557
3      3  90772459   93655574
4      4  49708101   51743951
5      5  46485901   50059807
6      6  58553889   59829934
7      7  58169654   60828234
8      8  44033745   45877265
9      9  43236168   45518558
10    10  39686683   41593521
11    11  51078349   54425074
12    12  34769408   37185252
13    13  16000001   18051248
14    14  16000001   18173523
15    15  17000001   19725254
16    16  36311159   38280682
17    17  22813680   26885980
18    18  15460900   20861206
19    19  24498981   27190874
20    20  26436233   30038348
21    21  10864561   12915808
22    22  12954789   15054318
X      X  58605580   62412542
Y      Y  10316945   10544039"""

    centromereshg38 = pd.DataFrame([c.split()[-3:] for c in centromereshg38.split('\n')],
                                   columns= "chrom left_base right_base".split())
    centromereshg38.index = centromereshg38.chrom.apply(lambda x: 'chr'+x)
    centromereshg38['left_base'] = centromereshg38.pop('left_base').apply(int)
    centromereshg38['right_base']=centromereshg38.pop('right_base').apply(int)

    genome = loadHumanGenome()
    centromereshg38['len'] = centromereshg38.apply(lambda x: len(genome.chromosomes[x.name]),axis=1)
    centromereshg38['qlen'] = centromereshg38.len - centromereshg38.right_base
    centromereshg38['chr_weight'] = centromereshg38.len/centromereshg38.len.max()
    centromereshg38['q_weight'] = centromereshg38.qlen/centromereshg38[['left_base','qlen']].max().max()
    centromereshg38['p_weight'] = centromereshg38.left_base/centromereshg38[['left_base','qlen']].max().max()
    del genome

    ensembl = get_ensembl()
    ensembl = ensembl[ensembl.chr.isin(centromereshg38.index)]
    ensembl['chrarm'] = ensembl.apply(lambda x: 'p' if x.stop < centromereshg38.ix[x.chr].left_base else
                                  ('q' if x.start > centromereshg38.ix[x.chr].right_base else 'pq'),axis=1)
    centromereshg38['chr_genes'] = ensembl.groupby('chr').size()
    centromereshg38['p_genes'] = ensembl[ensembl.chrarm=='p'].groupby('chr').size()
    centromereshg38['q_genes'] = ensembl[ensembl.chrarm=='q'].groupby('chr').size()
    centromereshg38['chr_gweight'] = centromereshg38.chr_genes/centromereshg38.chr_genes.max()
    centromereshg38['q_gweight'] = centromereshg38.q_genes/centromereshg38[['p_genes','q_genes']].max().max()
    centromereshg38['p_gweight'] = centromereshg38.p_genes/centromereshg38[['p_genes','q_genes']].max().max()
    del ensembl

    return centromereshg38

def get_msigdb():
    genesetsCollections = pickle.load(open('/home/christophe/Data/Genomes/MSigDB/msigdb.pickle','rb'))
    print('MSigDB {}'.format(genesetsCollections['version']))
    mdb = genesetsCollections['MSigDB']
    return mdb

#@retrieveSources -> not working login required
def get_msigdb6():
    """
    Source: http://software.broadinstitute.org/gsea/msigdb/download_file.jsp?filePath=/resources/msigdb/5.2/msigdb_v6.0.xml
    """
    import xml.etree.ElementTree as ET
    import pickle
    parser = ET.parse(processedDataStorage+'msigdb_v6.0.xml')
    root = parser.getroot()
    genesetsCollections = {} 
    for geneset in root:
        if (geneset.attrib['CATEGORY_CODE'] == 'ARCHIVED' or
            geneset.attrib['ORGANISM'] != 'Homo sapiens'): continue
        try:
            genesetsCollections[geneset.attrib['CATEGORY_CODE']
            ][geneset.attrib['STANDARD_NAME']] = geneset.attrib['MEMBERS_SYMBOLIZED'].split(',')
        except KeyError as e:
            genesetsCollections[geneset.attrib['CATEGORY_CODE']] = {}
            genesetsCollections[geneset.attrib['CATEGORY_CODE']
            ][geneset.attrib['STANDARD_NAME']] = geneset.attrib['MEMBERS_SYMBOLIZED'].split(',')
    genesetsCollections = {'version':'6','MSigDB':genesetsCollections}        
    print('MSigDB {}'.format(genesetsCollections['version']))
    mdb = genesetsCollections['MSigDB']
    return mdb
   
if __name__ == '__main__':
    import argparse
    from rpy2.rinterface import RRuntimeError
    from rpy2.robjects.packages import importr
    import rpy2.robjects as ro
    #Activate automatic pandas/r conversion
    from rpy2.robjects import pandas2ri
    pandas2ri.activate()

    parser = argparse.ArgumentParser(description='Generate LSD dataset and store as RData.')
    parser.add_argument('datasets', type=str, nargs='*', help='datasets to generate')
    parser.add_argument('--list', action='store_true', help='list available datasets')

    args = parser.parse_args()

    if args.list:
        listLSDatasets()
    



#!/usr/bin/env python
# Functions for linking survival to genes

from lifelines import KaplanMeierFitter
from lifelines.statistics import logrank_test
import matplotlib.pyplot as plt

# Calculate survival impact for each gene
def geneImpactSurvival(gene,expressions,metadata,groupingByQuantile=0.5,grouping=None,filter=None,
                       metacensorcol="overall_survival",metaDFDcol="death_from_disease",
                       plot=False,rounding=2):
    """
    Returns None if no meaningful groups
    """
    if filter is not None:
        expressions = expressions[expressions.columns[filter]]
        metadata = metadata[filter]
    if grouping is None:
        cutoff = expressions.ix[gene].quantile(groupingByQuantile) if groupingByQuantile else expressions.ix[gene].mean()
        groupHigh = expressions.ix[gene] > cutoff
    else: groupHigh = grouping

    kmf = KaplanMeierFitter()
    
    try: kmf.fit(metadata[metacensorcol][~groupHigh], metadata[metaDFDcol][~groupHigh], label='low')
    except ValueError: return None
    lastlow = float(kmf.survival_function_.ix[kmf.survival_function_.last_valid_index()])
    if plot:
        if type(plot) is bool:
            ax = kmf.plot()
        else: ax = kmf.plot(ax=plot)

    try: kmf.fit(metadata[metacensorcol][groupHigh], metadata[metaDFDcol][groupHigh], label='high')
    except ValueError: return None
    lasthigh = float(kmf.survival_function_.ix[kmf.survival_function_.last_valid_index()])
    if plot: kmf.plot(ax=ax)

    results = logrank_test(metadata[metacensorcol][groupHigh], metadata[metacensorcol][~groupHigh],
                           metadata[metaDFDcol][groupHigh], metadata[metaDFDcol][~groupHigh], alpha=.99)
    #results.print_summary()
    if not rounding:
        result = (lastlow-lasthigh,results.p_value)
    else:
        result = (round(lastlow-lasthigh,rounding),round(results.p_value,rounding))
    if plot: return result,ax
    else: return result

def geneCombinationImpactSurvival(genes,expressions,metadata,groupingByQuantile=0.5,
                       metacensorcol="overall_survival",metaDFDcol="death_from_disease",
                       plot=False,rounding=2):
    from itertools import combinations
    cutoff = {gene:expressions.ix[gene].quantile(groupingByQuantile) if groupingByQuantile else expressions.ix[gene].mean() for gene in genes}
    groupHigh = [expressions.ix[gene] > cutoff[gene] for gene in genes]
    kmf = KaplanMeierFitter()

    genis = [(i,s) for i in range(len(genes)) for s in ('H','L')]
    lastvalues = {}
    
    for combi in set(combinations(['H','L']*len(genes),len(genes))):
        selection = groupHigh[0] if combi[0] == 'H' else ~groupHigh[0]
        for gsel in zip(range(1,len(genes)),combi[1:]):
            selection = selection & (groupHigh[gsel[0]] if gsel[1] == 'H' else ~groupHigh[gsel[0]])
        if sum(selection) == 0: continue
        kmf.fit(metadata[metacensorcol][selection], metadata[metaDFDcol][selection], label=''.join(combi))
        lastvalues[combi] = (sum(selection),float(kmf.survival_function_.ix[kmf.survival_function_.last_valid_index()]))
        try: kmf.plot(ax=ax)
        except NameError: ax = kmf.plot()

    ax.set_title(':'.join(genes))
    return lastvalues

    #results = logrank_test(metadata[metacensorcol][groupHigh], metadata[metacensorcol][~groupHigh],
    #                       metadata[metaDFDcol][groupHigh], metadata[metaDFDcol][~groupHigh], alpha=.99)
    #results.print_summary()
    #if not rounding:
    #    return (lastlow-lasthigh,results.p_value)
    #else:
    #    return (round(lastlow-lasthigh,rounding),round(results.p_value,rounding))

def subsetsImpactSurvival(subsets,metadata,metacensorcol="overall_survival",
                          metaDFDcol="death_from_disease",plot=False,title=None,rounding=2):
    """
    subsets is a dictionary,
    e.g.: subsets={'cluster {}'.format(i):metadata.index.isin(fitrue.columns[kmeans.labels_==i]) for i in range(4)}
    """
    kmf = KaplanMeierFitter()

    lastvalues = {}
    for subset in subsets:
        kmf.fit(metadata[metacensorcol][subsets[subset]], metadata[metaDFDcol][subsets[subset]], label=subset)
        lastvalues[subset] = (sum(subsets[subset]),float(kmf.survival_function_.ix[kmf.survival_function_.last_valid_index()]))
        try: kmf.plot(ax=ax)
        except NameError: ax = kmf.plot()

    if title: ax.set_title(title)
    return lastvalues

# 2-way survival curves
def twoGeneSurvivalPlot(g1,g2,expressionData,metadata,ax=None):
    if not ax: f,ax = plt.subplots()
    g1med = expressionData.ix[g1].median()
    g2med = expressionData.ix[g2].median()
    gHH = expressionData.columns[(expressionData.ix[g1]>g1med)&(expressionData.ix[g2]>g2med)] #High g1, high g2
    gHL = expressionData.columns[(expressionData.ix[g1]>g1med)&(expressionData.ix[g2]<=g2med)]
    gLH = expressionData.columns[(expressionData.ix[g1]<=g1med)&(expressionData.ix[g2]>g2med)]
    gLL = expressionData.columns[(expressionData.ix[g1]<=g1med)&(expressionData.ix[g2]<=g2med)]
    kmf = KaplanMeierFitter()
    for g,a in zip((gHH,gHL,gLH,gLL),('HH','HL','LH','LL')):
        kmf.fit(metadata.lastfollowup[metadata.index.isin(g)],
                ~metadata.survived[metadata.index.isin(g)], label='{}[{}]'.format(a,len(g)))
        kmf.plot(ax=ax)
    ax.set_title('{} and {}'.format(g1,g2))    

def twoGeneSurvivalPlots(expressionData,metadata,genes=('BRIP1','TK1','BIRC5','EME1','WNT3')):
    from itertools import combinations, count
    numberOfPlots = len(list(combinations(range(len(genes)),2)))
    prows = int(np.ceil(numberOfPlots**0.5))
    pcols = int(np.floor(numberOfPlots**0.5))
    f,axes = plt.subplots(prows,pcols,sharex=True)
    c = count(0)
    for g1 in genes:
        for g2 in genes[genes.index(g1)+1:]:
            ax = axes.flatten()[next(c)]
            twoGeneSurvivalPlot(g1,g2,expressionData=expressionData,metadata=metadata,ax=ax)



#!/usr/bin/env python
import matplotlib.pyplot as plt
import matplotlib.patches as ptch
import seaborn as sns
import numpy as np, pandas as pd
import networkx as nx
from inspect import getmembers
from unittest.mock import MagicMock as Mock
from collections import OrderedDict
import LSD, itertools

#TODO http://bokeh.pydata.org/en/latest/docs/gallery.html#gallery

# General configuration
sns.set(style="whitegrid", palette="pastel", color_codes=True)

def drawGeneEnvNetwork(gene,interactome='string',addNeighborEdges=True,node_color='r',layout='random_layout'):
    """
    interactome -> one of the networks in LSD.get_proteinNetworks
    >> drawGeneEnvNetwork('BRIP1') # doctest: +ELLIPSIS
    <matplotlib.figure.Figure ...>
    """
    nws = LSD.get_proteinNetworks()
    nw = nws.__getattribute__(interactome+'nx')
    G = nx.Graph()
    neighbors = nw.neighbors(gene)
    [G.add_edge('BRIP1',n) for n in neighbors]
    if addNeighborEdges:
        for n in neighbors:
            nn = nw.neighbors(n)
            for nnn in nn:
                if nnn in neighbors: G.add_edge(n,nnn)
    fig,ax = plt.subplots()
    nx.draw_networkx(G,pos=drawGeneEnvNetwork.layouts[layout](G),
                     with_labels=True,
                     node_size=40,node_color=node_color,
                     ax=ax)
    ax.axis('off')
    return fig
drawGeneEnvNetwork.layouts = dict(getmembers(nx.layout))

def drawCNAcircos(cnaPositions,cnaTotal=False,chrRange=None,sortPositions=True,
                  genePositions=None,geneAnnotations=False,
                  startAngle=0,color='r',wedgebgshade='0.9',genecolor='k',ax=None):
    """
    color: either one color or a list of cnaPositions size

    Example BRIP1 on 17q
    >>> drawCNAcircos([(36094885,83084062),(59577514,83084062)],cnaTotal=10,chrRange=(26885980,83257441),
    ... genePositions={'BRIP1':61863521}) # doctest: +ELLIPSIS
    <matplotlib.figure.Figure ...>
    """
    if sortPositions:
        cnaPositions = sorted(cnaPositions,key=lambda x: max(x)-min(x),reverse=True)
    if not cnaTotal: cnaTotal = len(cnaPositions)
    wedgeDegrees = 360/cnaTotal
    maxChr,minChr = max(chrRange),min(chrRange)
    r = maxChr-minChr

    if ax: fig = ax.get_figure()
    else: fig,ax = plt.subplots()
    
    for cna in cnaPositions:
        if wedgebgshade: ax.add_patch(ptch.Wedge((0,0),r,startAngle,startAngle+wedgeDegrees,fc=wedgebgshade))
        wedge = ptch.Wedge((0,0),max(cna)-minChr,startAngle,startAngle+wedgeDegrees,width=max(cna)-min(cna),fc=color)#,ec=color)
        startAngle+=wedgeDegrees
        ax.add_patch(wedge)
    # Add circle edge for chr zoom
    ax.add_patch(ptch.Circle((0,0), radius=r, ec='k', fc='none'))
    # Add genes
    for g in genePositions:
        ax.add_patch(ptch.Circle((0,0), radius=genePositions[g]-minChr, ec=genecolor, fc='none'))
        if geneAnnotations == True: raise NotImplementedError

    ax.axis('off')
    ax.set_xlim((-r,r))
    ax.set_ylim((-r,r))
    return fig

def curvedHeatPlot(dataframe,columns,topDisplayed=10,cellwidth=.2,cellheight=.1,cmap='hot_r',
                   headingTextSize=14,curveLabels=True,filename=None):
    """
    >>> curvedHeatPlot(Mock(),['col1','col2']) # doctest: +ELLIPSIS
    <matplotlib.figure.Figure ...>
    """
    from itertools import count
    cmap = plt.get_cmap(cmap)
    topDisplayed_l_pos,topDisplayed_r_pos = topDisplayed,len(dataframe)-topDisplayed
    def curvedHeat(x,iterposition,ax,columns=columns):
        position = next(iterposition)
        if position < topDisplayed_l_pos:
            for c in columns:
                ax.add_patch(ptch.Rectangle((-1+cellwidth*columns.index(c),(topDisplayed_l_pos - position - 1)*cellheight),
                                            width=cellwidth,height=cellheight,color=cmap(x[c])))
                if position == 0:
                    ax.annotate('{:.2f}'.format(x[c]),
                                (-1+cellwidth*columns.index(c)+cellwidth/2,((topDisplayed_l_pos - position - 1)*cellheight)+.05),
                                ha='center',va='center')
            ax.annotate(x.name,(-1,((topDisplayed_l_pos - position - 1)*cellheight)+.05),ha='right',va='center')
        elif position >= topDisplayed_r_pos:
            for c in columns:
                ax.add_patch(ptch.Rectangle((1-cellwidth-(cellwidth*columns.index(c)),(position-topDisplayed_r_pos)*cellheight),
                                            width=cellwidth,height=cellheight,color=cmap(x[c])))
                if position == 57:
                    ax.annotate('{:.2f}'.format(x[c]),
                                (1-cellwidth*columns.index(c)-cellwidth/2,((position-topDisplayed_r_pos)*cellheight)+.05),
                                ha='center',va='center')
            ax.annotate(x.name,(1,((position-topDisplayed_r_pos)*cellheight)+.05),ha='left',va='center')
        else:
            startAngle = 180+(180*(position - topDisplayed)/(topDisplayed_r_pos - topDisplayed))
            endAngle = 180+(180*(1+position - topDisplayed)/(topDisplayed_r_pos - topDisplayed))
            for c in columns:
                ax.add_patch(ptch.Wedge((0,0),1-(cellwidth*columns.index(c)),startAngle,endAngle,cellwidth,color=cmap(x[c])))
            if curveLabels: ax.annotate(x.name,(np.pi*(startAngle+endAngle)/360,1),xycoords='polar',
                                        ha='right' if position < len(dataframe)/2 else 'left',va='top',
                                        rotation=(startAngle+endAngle)/2 + (180 if position < len(dataframe)/2 else 0))
    figheatmap,ax = plt.subplots(figsize=(6,4))
    ax.set_xlim((-1.2,1.2))
    ax.set_ylim((-1.2,(topDisplayed+2)*cellheight))
    for c in columns:
        ax.annotate(c,(-1+cellwidth*columns.index(c)+cellwidth/2,topDisplayed*cellheight+.05),ha='center',size=headingTextSize)
        ax.annotate(c,(1-cellwidth*columns.index(c)-cellwidth/2,topDisplayed*cellheight+.05),ha='center',size=headingTextSize)

    ax.axis('off')
    c = count(0)
    dataframe.T.apply(curvedHeat,args=(c,ax))
    
    if filename: figheatmap.savefig(filename,transparent=True)
    return figheatmap

def dosageViolin(gene,dataset,ax=None,cntype='gain',risksToPlot=3):
    """
    >>> dosageViolin('BRIP1',dataset=Mock()) # doctest: +SKIP
    """
    ds = dataset
    genedosage = pd.DataFrame({'expression': ds.exprdata.ix[gene],
                               'cna': ds.geneCNA.ix[gene],
    }).dropna()
    genedosage['risk_status'] = (genedosage.T.apply(lambda x: 'high' if ds.metadata.ix[x.name].high_risk=='1' else 'low')
                                 if risksToPlot == 2 else
                                 genedosage.T.apply(lambda x: 'lowrisk' if ds.metadata.ix[x.name].high_risk == '0' else
                                                   'highrisk_amp' if ds.metadata.ix[x.name].mycn_status == '1' else 'highrisk_sc')
    )
    
    genedosage = genedosage[genedosage.cna!=('loss' if cntype=='gain' else 'gain')]
    groupsizes = genedosage.groupby(['cna','risk_status']).size()
    #print(groupsizes)
        
    #Plots
    viofig,violax = (ax.get_figure(),ax) if ax else plt.subplots()
    try: violax.set_title('{} ({})'.format(gene,ds.name))
    except AttributeError: violax.set_title(gene)
    
    if risksToPlot == 2:
        order=['normal',cntype]
        hue_order = ['low','high']
        combinations = list(itertools.product(order,hue_order))
        sns.violinplot(x='cna', y='expression', hue='risk_status', data=genedosage, split=True,
                       inner='points', palette={'high':'b','low':'y'},order=order, hue_order=hue_order, ax=violax)
    elif risksToPlot == 3:
        order = ['lowrisk','highrisk_sc','highrisk_amp']
        hue_order = ['normal',cntype]
        combinations = list(itertools.product(order,hue_order))
        sns.violinplot(x='risk_status', y='expression', hue='cna', data=genedosage, split=True, inner='points',
                       palette={'normal':'y',cntype:'b' if cntype=='loss' else 'r'}, order=order, hue_order=hue_order, ax=violax)
    sns.despine(left=True)
    
    #Annotate sizes
    for c,i in zip(combinations,range(len(combinations))):
        violax.annotate(str(groupsizes[c[::-1]]) if c[::-1] in groupsizes else '0',(int(i/2)+(0.2 if i%2 else -0.2),.5),
                        xycoords=('data','axes fraction'),ha='center')
    
    return viofig

def draw_cellcycle(annotations,totalCompartments=100,phases=OrderedDict([('G1',0),('S',45),('G2',65),('M',85)]),ax=None,innerAnnotRing=.5):
    """
    annotations => pd.Series with genes in index and peak time values

    >>> ax = draw_cellcycle()
    """
    xytransform = lambda r,t: (r*np.cos(2*np.pi*(360-t)/360),r*np.sin(2*np.pi*(360-t)/360))
    if ax:
        fig = ax.get_figure()
    else: fig,ax = plt.subplots()
    padding = .1
    innerRadius = .8
    ax.axis('off')
    ax.set_xlim((-1-padding,1+padding))
    ax.set_ylim((-1-padding,1+padding))
    ax.add_patch(ptch.Circle(xy=(0,0),radius=1,fill=False,lw=2))
    ax.add_patch(ptch.Circle(xy=(0,0),radius=innerRadius,fill=False,lw=2))
    compartments = np.arange(0,360,360/totalCompartments) - 90
    phaseStart = list(phases.values())
    for i,p in enumerate(phases):
        ax.add_patch(ptch.PathPatch(ptch.Path([xytransform(1,compartments[phases[p]]),
                                               xytransform(innerRadius,compartments[phases[p]])]),color='k',lw=5))
        anAngle = np.mean((compartments[phaseStart[i]],compartments[phaseStart[i+1]] if i+1 < len(phases) else 360-90))
        ax.annotate(p,xytransform(np.mean((1,innerRadius)),anAngle),va='center',ha='center')
    for p,grp in annotations.groupby(by=annotations):
        ax.add_patch(ptch.PathPatch(ptch.Path([xytransform(1,compartments[p]),
                                               xytransform(innerRadius,compartments[p])]),color='r',lw=3))
        ax.annotate(','.join(grp.index),xytransform((1+padding) if len(grp) == 1 else innerAnnotRing,compartments[p]),va='center',ha='center',rotation=180-compartments[p])
    return ax



#!/bin/env python3
#CVN set of functions for sequence analysis
#General imports
import numpy as np

def recomplement(dna):
    """
    Returns the complement of a DNA motive for regex searching
    """
    return dna.translate(recomplement.dict)[::-1]
try: recomplement.dict=str.maketrans('ACTGactg()[]{}','TGACtgac)(][}{')
except AttributeError: #py2#
    import string
    recomplement.dict=string.maketrans('ACTGactg()[]{}','TGACtgac)(][}{')

class DNA:
    def __init__(self,name,sequence):
        self.name = name
        self.sequence = sequence

    def __len__(self):
        return len(self.sequence)

    def __getitem__(self, key):
        return self.sequence[key]

    def __str__(self):
        return self.name
    
    def __repr__(self):
        return ('{} ({}...)'.format(self.name,self.sequence[:10]))

class DNAregion:
    def __init__(self,dna,region):
        self.dna = dna
        self.region = region

    def __str__(self):
        return self.dna[region]
    
    def __repr__(self):
        return self.dna[region]

class Genome:
    """
    A simple class to represent a genome.
    """
    def __init__(self,species,assembly,chromosomes):
        self.species = species
        self.assembly = assembly
        self.chromosomes = {}
        for ch in chromosomes:
            self.addChromosome(ch.name,ch)

    def addChromosome(self,name,sequence):
        self.chromosomes[name] = sequence

    def __len__(self):
        return len(self.chromosomes)

    def __getitem__(self, key):
        if key[1].step and key[1].step < 0:
            return recomplement(self.chromosomes[key[0]][
                key[1].start-1:key[1].stop:abs(key[1].step)])
        else:
            return self.chromosomes[key[0]][key[1].start-1:key[1].stop:key[1].step]

    def windowSlider(self,windowSize=100,overlapping=True):
        for ch in sorted(self.chromosomes):
            for i in range(0,len(self.chromosomes[ch]),1 if overlapping else windowSize):
                yield (ch,slice(i,i+windowSize))

def loadHumanGenome():
    from glob import glob
    files = glob('/home/christophe/Data/Genomes/chroms/chr??.fa')
    files += glob('/home/christophe/Data/Genomes/chroms/chr?.fa')
    chromosomes = []
    for f in files:
        f = open(f).readlines()
        chromosomes.append(DNA(f.pop(0).strip()[1:],''.join([l.strip() for l in f])))
    return Genome('human','GRCh38',chromosomes)

class PFM():
    def __init__(self,pfmfile):
        pfm = open(pfmfile).readlines()
        self.name = pfm.pop(0).strip().split().pop()
        pfm = [l.replace('[','').replace(']','').strip().split() for l in pfm]
        self.nucleotides = np.array([l.pop(0) for l in pfm])
        self.pfm = np.array([[int(i) for i in l] for l in pfm])
        self.pfm_norm = (self.pfm/self.pfm.sum(0)).round(3)

    def setMinOfMax(self,minimal=0.30):
        self.includeInMotive = (self.pfm/self.pfm.max(0))>minimal

    def generateMotive(self,minimal=0.30):
        self.setMinOfMax(minimal)
        return (r'['+
                r']['.join([''.join(self.nucleotides[self.includeInMotive[:,i]]) for i in range(self.includeInMotive.shape[1])])+
                r']')

# Gene annotation functions
def literatureLinkSearch(term,referenceTerm='quadruplex'):
    """
    Returns all pubmed ids for pubs containing both term and referenceTerm
    """
    import requests
    import xml.etree.ElementTree as ET
    try: r = requests.get('https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi',
                          params={'db':'pubmed','term':'{}[Title/Abstract] AND {}[Title/Abstract]'.format(referenceTerm,term)})
    except Exception:
        return np.nan
    root = ET.fromstring(r.text)
    return [i.text for i in root.findall('IdList/Id')]

# Signature enrichment test
def calcSignature(counts,up,down=None,annotation=None,ax=None,xax=None,grouping=None,g4sig=False,
                  medianLines=True,cortests=True,**kwargs):
    """
    - signature (up,down) should be provided as (a) pandas.Series
    - grouping should be of the form (grouping_func,markers_dict) -> e.g. (lambda x: x.split('_')[0],{'SH-SY5Y':'s','IMR32':'o'})
            grouping_func should provide as result all markers_dict keys
    - xax if used should be pd.Series with index corresponding to counts.columns
    """
    import pandas as pd
    if type(up) == str:
        if not 'signatures' in dir(testSignature):
            testSignature.signatures = pd.read_table('/home/christophe/Dropbiz/Lab/z_archive/Datasets/dnupGenesetsOfInterest.gmt',
                                                     index_col=0,header=None)
            testSignature.sigdescriptions = testSignature.signatures.pop(1)
        up = testSignature.signatures.ix[up].dropna()
    if annotation is not None:
        up = annotation[annotation.gene_label.isin(up)].index
    ranks = counts.rank()
    rankSum = ranks[ranks.index.isin(up)].sum()
    if g4sig:
        rankSum = calcGlobalG4sig(ranks,ensembl,rank=False) if type(g4sig) == bool else g4sig
    xax = xax.sort_values() if xax is not None else pd.Series(range(len(rankSum)),rankSum.index)
    if ax:
        rankRel = rankSum/rankSum.max()
        if grouping and down is None:
            for gname,grp in rankRel.groupby(grouping[0]):
                ax.scatter([xax.ix[i] for i in grp.index],grp,c='r',s=30,marker=grouping[1][gname])
        else: ax.scatter([xax.ix[i] for i in rankRel.index],rankRel,c='r',label='up ({})'.format(len(up)),**kwargs)
        ax.set_xticks(xax)
        ax.set_xticklabels(xax.index,rotation=-30,ha='left')
        padding = (max(xax)-min(xax))*0.01
        ax.set_xlim((min(xax)-padding,max(xax)+padding))
        if medianLines:
            ax.axvline(xax.median())
            ax.axhline(rankRel.median())
        if cortests:
            from scipy.stats import spearmanr,pearsonr,fisher_exact
            spearesults = spearmanr([xax.ix[i] for i in rankRel.index],rankRel)
            ax.text(.99,.99,
                    'Spearman: {:.2f} ({:.4f})\nPearson: {:.2f} ({:.4f})\nFisherET: {:.2f} ({:.4f})'.format(
                        *spearesults,
                        *pearsonr([xax.ix[i] for i in rankRel.index],rankRel),
                        *fisher_exact([[sum((xax<xax.median())&(rankRel>rankRel.median())),
                                        sum((xax>xax.median())&(rankRel>rankRel.median()))],
                                       [sum((xax<xax.median())&(rankRel<rankRel.median())),
                                        sum((xax>xax.median())&(rankRel<rankRel.median()))]],
                                      alternative = 'less' if spearesults[0]<0 else 'greater')
                    ),
                    transform=ax.transAxes,va='top',ha='right')
    if down is not None:
        if type(down) == str:
            down = testSignature.signatures.ix[down].dropna()
        if annotation is not None:
            down = annotation[annotation.gene_label.isin(down)].index
        ranksDown = counts.rank(ascending=False)
        rankDownSum = ranksDown[ranksDown.index.isin(down)].sum()
        rankTotalSum = pd.DataFrame({'up':rankSum,'down':rankDownSum}).mean(axis=1)
        if ax:
            ax.scatter([xax.ix[i] for i in rankDownSum.index],rankDownSum/rankDownSum.max(),c='b',label='down ({})'.format(len(down)))
            if grouping:
                for gname,grp in (rankTotalSum/rankTotalSum.max()).groupby(grouping[0]):
                    ax.scatter([xax.ix[i] for i in grp.index],grp,c='g',s=40,marker=grouping[1][gname])
            else: ax.scatter([xax.ix[i] for i in rankTotalSum.index],rankTotalSum/rankTotalSum.max(),c='g',s=40,marker='s',label='mean')
            ax.legend()
        return rankTotalSum/rankTotalSum.max()
    else:
        return rankSum/rankSum.max()

def calcGlobalG4sig(countRanks,geneG4annotation,colG4='G4s',rank=True):
    """
    If rank=True, assume counts were given and rank them
    Scores like this -> sum(log(geneRank**G4s)) == sum(G4s*log(geneRank))
    """
    countRanks = countRanks[countRanks.index.isin(geneG4annotation.index)]
    if rank: countRanks = countRanks.rank()
    #countRanks.apply(lambda x: x.apply(lambda y,x=x: y**int(ensembl.ix[x.name].G4s)),axis=1).applymap(np.log).sum()
    return countRanks.applymap(
        np.log).apply(
            lambda x: x.apply(
                lambda y,x=x: int(geneG4annotation.ix[x.name][colG4])*y),axis=1).sum()    
    
    
# In main section below, some G4 'nifty' programming -> should be moved to dedicated G4 research script
if __name__ == '__main__':
    genome = loadHumanGenome()
    import re
    import matplotlib.pylab as plt
    q4m = r'.{1,7}'.join((r'G{3,7}' for g in range(2)))
    q4m_AT = q4m.replace('.','[AT]')
    q4motif = re.compile(q4m, re.IGNORECASE)
    q4motif_complement = re.compile(q4m.replace('G','C'), re.IGNORECASE)
    
    allowG4bulges = False
    if allowG4bulges:
        q4motif,q4motif_complement = (re.compile(r'|'.join([q4motif.pattern[:i*12]+
                                                            bulge+q4motif.pattern[i*12+6:]
                                                            for bulge in (r'GG.G',r'G.GG')
                                                            for i in range(4)]) , re.IGNORECASE)
                                      for q4motif in (q4motif,q4motif_complement))
    
    import random
    windowSize=1000
    binSize=50
    includeRandomizedWindows = False
    ATcontent = []
    G4content = []
    rG4content = []
    for s in genome.windowSlider(windowSize=windowSize,overlapping=False):
        ATcontent.append(genome[s].upper().count('A')+genome[s].upper().count('T'))
        G4content.append(len(q4motif.findall(genome[s]))+len(q4motif_complement.findall(genome[s])))
        if includeRandomizedWindows:
            seq = list(genome[s])
            random.shuffle(seq)
            seq = ''.join(seq)
            rG4content.append(len(q4motif.findall(seq))+len(q4motif_complement.findall(seq)))
            
    G4perATcontent = [[] for i in range(windowSize+1)]
    for a,g in zip(ATcontent,G4content):
        G4perATcontent[a].append(g)
    if includeRandomizedWindows:
        rG4perATcontent = [[] for i in range(windowSize+1)]
        for a,g in zip(ATcontent,rG4content):
            rG4perATcontent[a].append(g)
        
    #G4perATcontent_filtered = [l for l in G4perATcontent if len(l) > 30]
    #plt.violinplot(G4perATcontent_filtered[::15])
    G4perATbin=[]
    ticks = [100*(bin+(bin+binSize))/(2*windowSize) for bin in range(binSize,windowSize,binSize)]
    for bin in range(binSize,windowSize,binSize):
        l = []
        for i in G4perATcontent[bin:bin+binSize]: l+=i
        G4perATbin.append(l if l else [0])
    if includeRandomizedWindows:
        rG4perATbin=[]
        for bin in range(binSize,windowSize,binSize):
            l = []
            for i in rG4perATcontent[bin:bin+binSize]: l+=i
            rG4perATbin.append(l if l else [0])
    
    #plt.violinplot(G4perATbin,positions=ticks)
    plt.violinplot(G4perATbin)
    plt.title('G4s per AT content in {} bp windows'.format(windowSize))
    plt.xlabel('AT content (%)')
    plt.ylabel('# G4 (min GG.G tracts)')
    plt.xticks(range(1,len(ticks)+1),ticks)
    for i in range(len(G4perATbin)):
        g4, counts = np.unique(G4perATbin[i],return_counts=True)
        plt.scatter([i+1]*g4.shape[0],g4,s=(1+30*np.log10(counts)),c=counts,alpha=0.7)
    
    plt.show(block=False)    



#!/usr/bin/env python
# fegnome => Fisher exact genome module
# functions and classes for FE testing in relation to genomic context

#TODO https://www.math.hmc.edu/funfacts/ffiles/10006.3.shtml -> proportional volume
#TODO http://pythonhosted.org/gseapy/gseapy_example.html#prerank-example

import pandas as pd
import numpy as np
from collections import namedtuple, OrderedDict
import matplotlib.pyplot as plt
import matplotlib.patches as ptch
from scipy.stats import fisher_exact
from LSD import get_centromeres

centromereshg38 = get_centromeres()

def enrichometer(ranks,genesUp,genesDown=None,universe=None,fexact_H1='two-sided',
                 reservoirR1=1.4,reservoirR2=None,fillingLinewidths=False,ax=None,title=None,
                 feminpv=.05,invertx=False,textrotation=0,fontsize=12,**kwargs):
    """
    ranks -> pd.Series with ranked statistic
    if genesDown, is not provided, considered as unitary geneset
    universe -> int or list
     int that is the total number of genes in the universe
     or list with all genes in the universe 
    feminpv -> plot where enrichment is the strongest
     provide as float, if fisher enrichment at strongest is not
     significant in respect to value provided nothing is plotted
    """
    #enrichometer settings
    axiscale = 10
    eventoffset = 0
    if reservoirR1 and reservoirR2:
        eventlen = np.pi*reservoirR1*reservoirR2*len(ranks)/(axiscale*len(genesUp))
    else: eventlen = 1
    reservoirR2 = reservoirR2 if reservoirR2 else .8
    reservoirR1 = (reservoirR1 if reservoirR1 else
                   axiscale*eventlen*len(genesUp)/
                   (np.pi*reservoirR2*len(ranks)))
    padding = .1
    if not ax:
        fig,ax = plt.subplots(figsize=(8,2))
    else: fig = ax.get_figure()
    ax.axis('off')
    if fillingLinewidths:
        raise NotImplemented #todo use rectangle patches, does not work with linewidths
        axPixels = ax.transData.transform([(0,1),(1,0)])-ax.transData.transform((0,0))
        linewidths = (axPixels[1,0]/fig.dpi)*axiscale/len(ranks)
    settings = {
        'colors':'r',
        }
    settings.update(kwargs)

    #Normalize the ranks
    minR,maxR = (ranks.min(),ranks.max())
    normalized = ((ranks-minR)*axiscale/(maxR-minR)-axiscale-reservoirR1)
    
    ax.eventplot(normalized[normalized.index.isin(genesUp)],
                 lineoffsets=eventoffset,linelengths=eventlen,**settings)
    ax.set_xlim((-axiscale-reservoirR1-padding,reservoirR1+padding))
    ax.set_ylim((eventoffset-max(eventlen/2,reservoirR2)-padding,eventoffset+max(eventlen/2,reservoirR2)+padding))

    #Draw thermometer
    overlap = sum(ranks.index.isin(genesUp))
    filled = 1-overlap/len(genesUp)
    emptyReservoir = reservoirR1*2-(reservoirR1*2*filled)
    ax.add_patch(ptch.Ellipse((0,eventoffset),width=2*reservoirR1,height=2*reservoirR2,facecolor='r',edgecolor='none'))   
    ax.add_patch(ptch.Rectangle((-reservoirR1,eventoffset-reservoirR2),
                                emptyReservoir,2*reservoirR2,facecolor='w',edgecolor='none'))
    ax.add_patch(ptch.Ellipse((0,eventoffset),width=2*reservoirR1,height=2*reservoirR2,facecolor='none',edgecolor='k'))
    ax.add_patch(ptch.Rectangle((-axiscale-reservoirR1,eventoffset-eventlen/2),axiscale,eventlen,facecolor='w',edgecolor='k'))

    #Annotations
    ax.annotate('{:g}'.format(minR),(-axiscale-reservoirR1,(eventoffset-eventlen/2)-padding),
                ha='right' if invertx else 'left',va='top',rotation=textrotation,size=fontsize)
    ax.annotate('{:g}'.format(maxR),(-reservoirR1,(eventoffset-eventlen/2)-padding),
                ha='left' if invertx else 'right',va='top',rotation=textrotation,size=fontsize)

    if universe:
        if type(universe) is list: raise NotImplemented
        odds,pval = fisher_exact([[overlap,len(ranks)],[len(genesUp)-overlap,universe]],alternative=fexact_H1)
        ax.annotate('{}\n{:.3g}'.format(len(genesUp),pval),(0,eventoffset),
                    ha='center',va='center',rotation=textrotation,size=fontsize)

    if feminpv:
        fenrichscores = fenrichmentscore(ranks,genesUp)
        pvmin = fenrichscores.pvalue.min()
        if pvmin <= feminpv:
            leadingEdgeGene = fenrichscores[fenrichscores.pvalue==pvmin].first_valid_index()
            #ax.eventplot((normalized.ix[leadingEdgeGene],),lineoffsets=eventoffset,linelengths=eventlen,
            #             color='g')
            ax.add_patch(ptch.Rectangle((-axiscale-reservoirR1,eventoffset-eventlen/2),
                                        normalized.ix[leadingEdgeGene]+axiscale+reservoirR1,
                                        eventlen,facecolor='r',alpha=.4))
            ax.annotate('{:.3g}'.format(pvmin),(normalized.ix[leadingEdgeGene]+padding,eventoffset),
                        ha='left',va='center',size=fontsize)
        else: leadingEdgeGene = None
        
    if title:
        #ax.set_title(title)
        ax.annotate(title,(-axiscale-reservoirR1,eventoffset+padding+eventlen/2),
                    ha='right' if invertx else 'left',va='bottom',size=fontsize+2)

    if invertx: ax.invert_xaxis()
    
    return (fig,leadingEdgeGene) if feminpv else fig

def fenrichmentscore(ranks,genesUp,genesDown=None):
    """
    Calculate fe score for each position in the ranked list
    """
    genes = {g for g in genesUp if g in ranks.index}
    ranks = ranks.sort_values()
    fescores = []
    for i in range(1,len(ranks)):
        genesUpto,genesAfter = set(ranks.index[:i]),set(ranks.index[i:])
        fescores.append(
            fisher_exact([[len(genesUpto&genes),len(genesUpto-genes)],
                          [len(genesAfter&genes),len(genesAfter-genes)]],
                         alternative='greater')
        )
        
    return pd.DataFrame(fescores,columns=['oddratio','pvalue'],index=ranks.index[:-1])

RankSumResult = namedtuple('RankSumResult',[
    'probability',
    'genesetsize',
    'overlapsize',
    'rankedgenes',
    'nulldistrosize',
    'fe_s','fe_p'])
        
def rankSumProbability(ranks,geneset,alternative='greater',nulldistrosize=10000,nulldistrosave=True,universe=30000,verbose=False):
    """
    ranks: pd.Series
    geneset: set

    >>> from .tests.test_fegnome import testRanks, testGeneset
    >>> rankSumProbability(testRanks,testGeneset,nulldistrosave=False) # doctest: +ELLIPSIS
    RankSumResult(...)
    """
    realsetlen = len(geneset)
    setlen = ranks.index.isin(geneset).sum()
    if not setlen: return RankSumResult(None,realsetlen,setlen,None,None,None,None)
    ranksum = ranks[ranks.index.isin(geneset)].sum()
    if nulldistrosave:
        import pickle,hashlib,LSD
        hashindex = hashlib.md5(str(tuple(sorted(ranks))).encode()).hexdigest()
        try: nulldistros = pickle.load(open(LSD.processedDataStorage+'ranklogsum_nulldistros.pickle','rb'))
        except FileNotFoundError: nulldistros = {}
        if (hashindex,setlen) in nulldistros and nulldistros[(hashindex,setlen)][0] >= nulldistrosize:
            nulldistrosize,nulldistro = nulldistros[(hashindex,setlen)]
            if verbose: print('Using previously stored nulldistribution size',nulldistrosize)
        else:
            nulldistro = pd.Series([ranks.sample(n=setlen).sum() for i in range(nulldistrosize)])
            nulldistros[(hashindex,setlen)] = (nulldistrosize, nulldistro)
            pickle.dump(nulldistros,open(LSD.processedDataStorage+'ranklogsum_nulldistros.pickle','wb'))
    else:
        nulldistro = pd.Series([ranks.sample(n=setlen).sum() for i in range(nulldistrosize)])

    if alternative == 'greater':
        probability = float(sum(nulldistro < ranksum))/nulldistrosize
    elif alternative == 'less':
        probability = float(sum(nulldistro > ranksum))/nulldistrosize

    fe = fisher_exact([[setlen,len(ranks)],
                       [realsetlen-setlen,universe]],
                      alternative='greater')
        
    return RankSumResult(probability,realsetlen,setlen,len(ranks),nulldistrosize,*fe)



from setuptools import setup

setup(name = 'bidali',
      version = '0.1.0',
      description = 'Biological Data Analysis Library',
      url = 'https://github.com/beukueb/bidali',
      author = 'Christophe Van Neste',
      author_email = 'christophe.vanneste@ugent.be',
      license = 'MIT',
      packages = ['bidali','LSD'],
      install_requires = [
          #Generated with `pipreqs .` and then moved here from requirements.txt
          'gffutils==0.8.7.1',
          'scipy==0.19.0',
          'lifelines==0.9.4',
          'seaborn==0.7.1',
          'requests==2.13.0',
          'pyliftover==0.3',
          'networkx==1.11',
          'rpy2==2.8.5',
          'matplotlib==2.0.0',
          'pandas==0.19.2',
          'setuptools==32.2.0',
          'numpy==1.12.0',
      ],
      zip_safe = False,
      #entry_points = {
      #    'console_scripts': ['getLSDataset=LSD.command_line:main'],
      #},
      test_suite = 'nose.collector',
      tests_require = ['nose']
)

#To install with symlink, so that changes are immediately available:
#pip install -e . 



if __name__ == '__main__':
    import argparse, inspect, LSD
    from rpy2.rinterface import RRuntimeError
    from rpy2.robjects.packages import importr
    import rpy2.robjects as ro
    #Activate automatic pandas/r conversion
    from rpy2.robjects import pandas2ri
    pandas2ri.activate()

    parser = argparse.ArgumentParser(description=
            '''
Generate LSD dataset and store as RData.
datasets have to be provided with their full path e.g. LSD.get_centromeres
            '''
    )
    parser.add_argument('datasets', type=str, nargs='*', help='datasets to generate')
    parser.add_argument('--list', action='store_true', help='list available datasets')

    args = parser.parse_args()

    if args.list:
        LSD.listLSDatasets()

    for ds in args.datasets:
        if not ds.startswith('LSD.'):
            raise Exception('Not an LSD dataset. `'+ds+'`should start with `LSD.`')
        datasetname = ds.split('.')[-1].replace('get_','')
        filename_r = '{}{}.RData'.format(LSD.processedDataStorage,datasetname)

        # Generate/load dataset
        m = inspect.importlib.import_module(ds[:ds.rindex('.')])
        f = m.__dict__[ds.split('.')[-1]]
        ds = f()
        ds_r = ro.ListVector(
            {d:ds.__getattribute__(d) for d in ds.__datasets__}
        )
        ro.globalenv[datasetname] = ds_r
        ro.r.save(datasetname,file=filename_r)
        print('{} now available for use in R with `load("{}")`'.format(datasetname,filename_r))
